/*
 *   Emu42.h
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "types.h"

#define	ARRAYSIZEOF(a)	(sizeof(a) / sizeof(a[0]))

#define	_B(a)			((a)*2)
#define	_KB(a)			((a)*2*1024)

#define HARDWARE		"Lewis"				// emulator hardware

// valid calculator models
#define MODELS			(MODELS_SACA MODELS_LEWIS)

#define MODELS_SACA		_T("ILNP")			// valid calculator models for Sacajawea hardware
#define MODELS_LEWIS	_T("DMOTUY")		// valid calculator models for Lewis hardware

#define BINARYHEADER28S "HPHP28-S"

#define RAW_FILTER		"User Code Files (*.RAW)\0*.RAW\0All Files (*.*)\0*.*\0"
#define HP_FILTER		"HP Binary Object (*.HP)\0*.HP\0All Files (*.*)\0*.*\0"

#define T2CYCLES		(dwCyclesRate)		// CPU cycles in 16384 Hz time frame

#define SM_RUN			0					// states of cpu emulation thread
#define SM_INVALID		1
#define SM_RETURN		2
#define SM_SLEEP		3

#define S_ERR_NO        0					// stack errorcodes
#define S_ERR_OBJECT	1
#define S_ERR_BINARY    2
#define S_ERR_ASCII     3

#define BAD_OB			(0xFFFFFFFF)		// bad object

#define MACRO_OFF		0					// macro recorder off
#define MACRO_NEW		1
#define MACRO_PLAY		2

#define TOPBAR_OFF		0x1					// top bar off
#define MENUBAR_OFF		0x2					// menu bar off

// MMU bit settings
#define PAGE_BITS		10					// valid bits for the page
#define ADDR_BITS		(20-PAGE_BITS)		// valid bits for the address
#define ADDR_SIZE		(1<<ADDR_BITS)		// address size (in nibbles)

// macro to check for valid calculator model
#define isModelValid(m)	(m != 0 && _tcschr(MODELS,      m) != NULL)
#define isModelLewis(m)	(m != 0 && _tcschr(MODELS_LEWIS,m) != NULL)
#define isModelSaca(m)	(m != 0 && _tcschr(MODELS_SACA, m) != NULL)

// values for emulator hardware platform
enum HDW_PLATFORM { HDW_UNKNOWN = 0, HDW_LEWIS, HDW_SACA };

// values for selection of Lewis chip
enum CHIP { MASTER = 0, SLAVE };

// values for mapping area
enum MMUMAP { M_NONE,
			  M_REG,  M_DISP,  M_NCE2,  M_NCE3,  M_ROM,
			  M_REGS, M_DISPS, M_NCE2S, M_NCE3S, M_ROMS
			};

// Emu42.c
extern HPALETTE			hPalette;
extern HPALETTE			hOldPalette;
extern HANDLE			hEventShutdn;
extern LPTSTR			szTitle;
extern CRITICAL_SECTION	csGDILock;
extern CRITICAL_SECTION	csKeyLock;
extern CRITICAL_SECTION	csIOLock;
extern CRITICAL_SECTION	csT1Lock;
extern CRITICAL_SECTION	csT2Lock;
extern CRITICAL_SECTION csSlowLock;
extern INT				nArgc;
extern LPCTSTR			*ppArgv;
extern LARGE_INTEGER	lFreq;
extern LONG				lrcTop;
extern HINSTANCE		hApp;
extern HWND				hWnd;
extern HDC				hWindowDC;
extern UINT             nAppZoom;
extern DWORD			dwWakeupDelay;
extern BOOL				bAutoSave;
extern BOOL				bAutoSaveOnExit;
extern BOOL				bSaveDefConfirm;
extern BOOL				bStartupBackup;
extern BOOL				bAlwaysDisplayLog;
extern BOOL				bLoadObjectWarning;
extern HANDLE			hThread;
extern DWORD			lThreadId;
extern VOID				SetWindowTitle(LPCTSTR szString);

// mru.c
extern BOOL    MruInit(INT nNum);
extern VOID    MruCleanup(VOID);
extern VOID    MruAdd(LPCTSTR lpszEntry);
extern VOID    MruRemove(INT nIndex);
extern VOID    MruMoveTop(INT nIndex);
extern INT     MruEntries(VOID);
extern LPCTSTR MruFilename(INT nIndex);
extern VOID    MruUpdateMenu(HMENU hMenu);
extern VOID    MruWriteList(VOID);
extern VOID    MruReadList(VOID);

// Settings.c
extern VOID ReadSettings(VOID);
extern VOID WriteSettings(VOID);
extern VOID ReadLastDocument(LPTSTR szFileName, DWORD nSize);
extern VOID WriteLastDocument(LPCTSTR szFilename);
extern VOID ReadSettingsString(LPCTSTR lpszSection, LPCTSTR lpszEntry, LPCTSTR lpDefault, LPTSTR lpData, DWORD dwSize);
extern VOID WriteSettingsString(LPCTSTR lpszSection, LPCTSTR lpszEntry, LPTSTR lpData);
extern INT  ReadSettingsInt(LPCTSTR lpszSection, LPCTSTR lpszEntry, INT nDefault);
extern VOID WriteSettingsInt(LPCTSTR lpszSection, LPCTSTR lpszEntry, INT nValue);
extern VOID DelSettingsKey(LPCTSTR lpszSection, LPCTSTR lpszEntry);

// Display.c
extern UINT   nFullScreen;
extern UINT   nVertical;
extern UINT   nBackgroundX;
extern UINT   nBackgroundY;
extern UINT   nBackgroundW;
extern UINT   nBackgroundH;
extern UINT   nLcdX;
extern UINT   nLcdY;
extern UINT   nLcdZoom;
extern UINT   nLcdyZoom;
extern UINT   nLcdXOff;
extern UINT   nLcdYOff;
extern UINT   nLcdXPixel;
extern UINT   nLcdYPixel;
extern UINT   nLcdDistance;
extern HDC    hLcdDC;
extern HDC    hMainDC;
extern VOID   (*UpdateContrast)(VOID);
extern VOID   SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue);
extern VOID   GetSizeLcdBitmap(INT *pnX, INT *pnY, INT *pnRot180);
extern VOID   CreateLcdBitmap(VOID);
extern VOID   DestroyLcdBitmap(VOID);
extern BOOL   CreateMainBitmap(LPCTSTR szFilename);
extern VOID   DestroyMainBitmap(VOID);
extern BOOL   CreateAnnunBitmap(LPCTSTR szFilename);
extern VOID   DestroyAnnunBitmap(VOID);
extern VOID   StartDisplay(VOID);
extern VOID   StopDisplay(VOID);
extern VOID   UpdateAnnunciators(VOID);
extern VOID   ResizeWindow(VOID);

// Engine.c
extern CHIPSET_M Chipset;
extern CHIPSET_S ChipsetS;
extern BOOL      bInterrupt;
extern UINT      nState;
extern UINT      nNextState;
extern BOOL      bRealSpeed;
extern BOOL      bKeySlow;
extern UINT      nOpcSlow;
extern DWORD     dwSacaCycles;
extern DWORD     dwLewisCycles;
extern DWORD     dwCyclesRate;
extern VOID      InitAdjustSpeed(VOID);
extern VOID      AdjKeySpeed(VOID);
extern VOID      SetSpeed(BOOL bAdjust);
extern BOOL      WaitForSleepState(VOID);
extern UINT      SwitchToState(UINT nNewState);
extern UINT      WorkerThread(LPVOID pParam);

// Fetch.c
extern VOID    EvalOpcode(LPBYTE I);

// Files.c
extern TCHAR   szEmuDirectory[MAX_PATH];
extern TCHAR   szDocDirectory[MAX_PATH];
extern TCHAR   szCurrentKml[MAX_PATH];
extern TCHAR   szBackupKml[MAX_PATH];
extern TCHAR   szCurrentFilename[MAX_PATH];
extern TCHAR   szBackupFilename[MAX_PATH];
extern TCHAR   szBufferFilename[MAX_PATH];
extern BOOL    bDocumentAvail;
extern LPBYTE  mNCE2;
extern LPBYTE  mNCE3;
extern LPBYTE  sNCE2;
extern LPBYTE  sNCE3;
extern LPBYTE  pbyRom;
extern DWORD   dwRomSize;
extern WORD    wRomCrc;
extern UINT    nCurrentHardware;
extern BYTE    cCurrentRomType;
extern UINT    nCurrentClass;
extern BOOL    bBackup;
extern LPCTSTR FullFilename(LPCTSTR lpszFilename);
extern BOOL    PatchRom(LPCTSTR szFilename);
extern BOOL    CrcRom(WORD *pwChk);
extern BOOL    MapRom(LPCTSTR szFilename);
extern VOID    UnmapRom(VOID);
extern VOID    ResetDocument(VOID);
extern BOOL    NewDocument(VOID);
extern BOOL    OpenDocument(LPCTSTR szFilename);
extern BOOL    SaveDocument(VOID);
extern BOOL    SaveDocumentAs(LPCTSTR szFilename);
extern BOOL    SaveBackup(VOID);
extern BOOL    RestoreBackup(VOID);
extern BOOL    ResetBackup(VOID);
extern VOID    SaveLastDocumentFolder(LPCTSTR);
extern BOOL    GetOpenFilename(VOID);
extern BOOL    GetSaveAsFilename(VOID);
extern BOOL    GetLoadObjectFilename(LPCTSTR lpstrFilter);
extern BOOL    GetSaveObjectFilename(LPCTSTR lpstrFilter);
extern WORD    WriteStack(UINT nStkLevel,LPBYTE lpBuf,DWORD dwSize);
extern BOOL    LoadObject(LPCTSTR szFilename);
extern BOOL    SaveObject(LPCTSTR szFilename);
extern HBITMAP LoadBitmapFile(LPCTSTR szFilename);

// Timer.c
extern VOID  SetHPTime(VOID);
extern VOID  StartTimers(enum CHIP eChip);
extern VOID  StopTimers(enum CHIP eChip);
extern DWORD ReadT2(enum CHIP eChip);
extern VOID  SetT2(enum CHIP eChip, DWORD dwValue);
extern BYTE  ReadT1(enum CHIP eChip);
extern VOID  SetT1(enum CHIP eChip, BYTE byValue);

// Mops.c
extern LPBYTE      RMap[1<<PAGE_BITS];
extern LPBYTE      WMap[1<<PAGE_BITS];
extern WORD        UpCRC(WORD wCRC, BYTE nib);
extern VOID        InitIO(VOID);
extern VOID        Map(WORD a, WORD b);
extern VOID        Config(VOID);
extern VOID        Uncnfg(VOID);
extern enum MMUMAP MapData(DWORD d);
extern VOID        Reset(VOID);
extern VOID        C_Eq_Id(VOID);
extern VOID        CpuReset(VOID);
extern VOID        Npeek(BYTE *a, DWORD d, UINT s);
extern VOID        Nread(BYTE *a, DWORD d, UINT s);
extern VOID        Nwrite(BYTE *a, DWORD d, UINT s);
extern BYTE        Read2(DWORD d);
extern DWORD       Read5(DWORD d);
extern VOID        Write5(DWORD d, DWORD n);
extern VOID        Write2(DWORD d, BYTE n);
extern VOID        IOBit(BYTE *d, BYTE b, BOOL s);

// Lowbat.c
extern BOOL bLowBatDisable;
extern VOID StartBatMeasure(VOID);
extern VOID StopBatMeasure(VOID);
extern VOID GetBatteryState(BOOL *pbLBI, BOOL *pbVLBI);

// Keyboard.c
extern DWORD dwKeyMinDelay;
extern VOID  ScanKeyboard(BOOL bActive, BOOL bReset);
extern VOID  KeyboardEvent(BOOL bPress, UINT out, UINT in);

// Keymacro.c
extern INT     nMacroState;
extern INT     nMacroTimeout;
extern BOOL    bMacroRealSpeed;
extern DWORD   dwMacroMinDelay;
extern VOID    KeyMacroRecord(BOOL bPress, UINT out, UINT in);
extern LRESULT OnToolMacroNew(VOID);
extern LRESULT OnToolMacroPlay(VOID);
extern LRESULT OnToolMacroStop(VOID);
extern LRESULT OnToolMacroSettings(VOID);

// Stack.c
extern BOOL    bDetectClpObject;
extern LRESULT OnStackCopy(VOID);
extern LRESULT OnStackPaste(VOID);

// Usrprg32.c
extern BOOL GetUserCode32(LPCTSTR szUCFile);
extern BOOL OnSelectProgram32(VOID);

// Usrprg42.c
extern BOOL GetUserCode42(LPCTSTR szUCFile);
extern BOOL OnSelectProgram42(VOID);

// RPL.c
extern DWORD RPL_SkipOb(DWORD d);
extern DWORD RPL_ObjectSize(BYTE *o,DWORD s);
extern DWORD RPL_CreateTemp(DWORD l);
extern UINT  RPL_Depth(VOID);
extern DWORD RPL_Pick(UINT l);
extern VOID  RPL_Replace(DWORD n);
extern VOID  RPL_Push(UINT l,DWORD n);
extern VOID  RPL_PushX(DWORD n);

// External.c
extern DWORD dwWaveVol;
extern VOID  External0(CHIPSET_M* w);
extern VOID  External1(CHIPSET_M* w);

// Message Boxes
static __inline int InfoMessage(LPCTSTR szMessage)  {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONINFORMATION|MB_SETFOREGROUND);}
static __inline int AbortMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONSTOP|MB_SETFOREGROUND);}
static __inline int YesNoMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNO|MB_ICONEXCLAMATION|MB_SETFOREGROUND);}
static __inline int YesNoCancelMessage(LPCTSTR szMessage,UINT uStyle) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNOCANCEL|MB_ICONEXCLAMATION|MB_SETFOREGROUND|uStyle);}

// Missing Win32 API calls
static __inline LPTSTR DuplicateString(LPCTSTR szString)
{
	UINT   uLength = lstrlen(szString) + 1;
	LPTSTR szDup   = malloc(uLength*sizeof(szDup[0]));
	lstrcpy(szDup,szString);
	return szDup;
}

// Full screen dialog box
static __inline VOID FULLSCREEN(HWND hDlg,DWORD dwFlags)
{
	SHINITDLGINFO shidi;

	_ASSERT(hWnd != NULL);
	SHFullScreen(hWnd, SHFS_SHOWTASKBAR | SHFS_SHOWSIPBUTTON | SHFS_SHOWSTARTICON);

	shidi.dwMask = SHIDIM_FLAGS;
	shidi.dwFlags = SHIDIF_SIZEDLG | dwFlags;
	shidi.hDlg = hDlg;
	SHInitDialog(&shidi);
	return;
}

// WinCE replacement
static __inline DWORD GetCurrentDirectory(DWORD nBufferLength,LPTSTR lpBuffer)
{
	LPTSTR lpPos;

	// get path and name of current file
	GetModuleFileName(NULL,lpBuffer,nBufferLength);

	// get delimiter position of filename
	if ((lpPos = _tcsrchr(lpBuffer,_T('\\'))) != NULL)
	{
		*lpPos = 0;							// purge filename
	}
	else
	{
		lpBuffer[0] = _T('\\');				// root path
		lpBuffer[1] = 0;
	}
	return lstrlen(lpBuffer);
}
