/*
 *   external.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"

#define MUSIC_FREQ 11025					// this can be adjusted for quality

#define CPU_FREQ	524288					// base CPU frequency
#define CPU_FREQNOM	(CPU_FREQ * 2)			// nominal CPU frequency

DWORD dwWaveVol = 64;						// wave sound volume

static VOID BeepWave(DWORD dwFrequency,DWORD dwDuration)
{
	HWAVEOUT     hSoundDevice;
	WAVEFORMATEX wf;
	WAVEHDR      wh;
	HANDLE       hEventSound;
	DWORD        i;

	if (dwFrequency == 0)					// this is just a delay
	{
		Sleep(dwDuration);
		return;
	}

	hEventSound = CreateEvent(NULL,FALSE,FALSE,NULL);

	wf.wFormatTag      = WAVE_FORMAT_PCM;
	wf.nChannels       = 1;
	wf.nSamplesPerSec  = MUSIC_FREQ;
	wf.nAvgBytesPerSec = MUSIC_FREQ;
	wf.nBlockAlign     = 1;
	wf.wBitsPerSample  = 8;
	wf.cbSize          = 0;

	if (waveOutOpen(&hSoundDevice,WAVE_MAPPER,&wf,(DWORD)hEventSound,0,CALLBACK_EVENT) != 0)
	{
		CloseHandle(hEventSound);			// no sound available
		return;
	}

	// (samp/sec) * msecs * (secs/msec) = samps
	wh.dwBufferLength = (DWORD) ((QWORD) MUSIC_FREQ * dwDuration / 1000);
	VERIFY(wh.lpData = malloc(wh.dwBufferLength));
	wh.dwBytesRecorded = 0;
	wh.dwUser = 0;
	wh.dwFlags = 0;
	wh.dwLoops = 0;

	for (i = 0; i < wh.dwBufferLength; ++i)	// generate square wave
	{
		wh.lpData[i] = (BYTE) ((((QWORD) 2 * dwFrequency * i / MUSIC_FREQ) & 1) * dwWaveVol);
	}

	VERIFY(waveOutPrepareHeader(hSoundDevice,&wh,sizeof(wh)) == MMSYSERR_NOERROR);

	ResetEvent(hEventSound);				// prepare event for finishing
	VERIFY(waveOutWrite(hSoundDevice,&wh,sizeof(wh)) == MMSYSERR_NOERROR);
	WaitForSingleObject(hEventSound,INFINITE); // wait for finishing

	VERIFY(waveOutUnprepareHeader(hSoundDevice,&wh,sizeof(wh)) == MMSYSERR_NOERROR);
	VERIFY(waveOutClose(hSoundDevice) == MMSYSERR_NOERROR);

	free(wh.lpData);
	CloseHandle(hEventSound);
	return;
}

VOID External0(CHIPSET_M* w)				// =BPUTL patch for HP17BII and HP19BII
{
	enum MMUMAP e;

	DWORD dwF, dwD;
	DWORD freq,dur;

	dwD = Npack(w->C,3);					// get fdd control

	e = MapData(w->pc);						// MMU area of code
	if (M_ROM == e || M_ROMS == e)
		dwF = 79 + 54 * (dwD >> 8);			// F, internally (LEWIS ROM)
	else
		dwF = 107 + 66 * (dwD >> 8);		// F, MEMC (External ROM or RAM)
	dwD = dwF * (4096 - 16 * (dwD & 0xFF));	// D

	// CPU frequency
	freq = ((DWORD) w->IORam[RATECTL] + 1) * CPU_FREQ / 4;

	dur = (dwD * 1000) / freq;				// duration in ms
	freq = freq / (2 * dwF);				// frequency in Hz

	if (freq < 37)   freq = 37;				// low limit of freqency (NT)
	if (freq > 4400) freq = 4400;			// high limit of HP

	BeepWave(freq,dur);						// do it on the hard way

	w->cycles += dur * 1000;				// estimate cpu cycles for beeping time (1MHz)

	// original routine return with...
	w->P = 0;								// P=0
	w->carry = FALSE;						// RTNCC
	w->pc = rstkpop();
	return;
}

VOID External1(CHIPSET_M* w)				// =makebeep patch
{
	DWORD freq,dur,cpufreq;

	freq = Npack(w->D,5);					// frequency in Hz
	dur = Npack(w->C,5);					// duration in ms

	// CPU freqency
	cpufreq = ((DWORD) w->IORam[RATECTL] + 1) * CPU_FREQ / 4;

	// frequency and duration RATE dependend
	freq = (DWORD) (((QWORD) freq * cpufreq) / CPU_FREQNOM);
	dur  = (DWORD) (((QWORD) dur * CPU_FREQNOM) / cpufreq);

	if (freq < 37)   freq = 37;				// low limit of freqency (NT)
	if (freq > 4400) freq = 4400;			// high limit of HP

	BeepWave(freq,dur);						// do it on the hard way

	w->cycles += dur * 1000;				// estimate cpu cycles for beeping time (1MHz)

	// original routine return with...
	w->P = 0;								// P=0
	w->intk = TRUE;							// INTON
	w->carry = FALSE;						// RTNCC
	w->pc = rstkpop();
	return;
}
